//////////////////////////////////////////////////////////////////////
//	depMatrix.cpp:	Use a adjecent matrix to express the dependence	//
//					relation between two files						//
//	ver 1.1			revision on ver 1.0:							//
//					function adjust() was revised to meet the re-	//
//					quirement which ask for placing the header file	//
//					and	implementation file together if possible	//
//	Language:		Visual C++, ver 6.0								//
//	Platform:														//
//	Application:	CSE687 projet 1									//
//	Author:			Qiang Zheng, qzheng@syr.edu						//
//																	//
//	About the definition and the explanation, please see the		//
//	depMatrix.h														//
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//	Purpose:	To use an adjecent matrix to express the file		//
//		        dependecies. And to maintain this matrix.			//
//				And an assotiative array is used to associate		//
//				the index of the row of the	matrix with the file	//
//				Name.												//
//																	//
//	e.g. fileA.cpp-->fileB.h-->										//
//		       |              |										//
//  		    ----------->fileC.h---->fileD.h						//
//																	//
//		(Here "-->" means "dependent on")							//
//																	//
//---- The corresponding adjecent matrix is:						//
//---- (F,T,T,T														//
//----  F,F,F,T														//
//----  F,F,F,T														//
//----  F,F,F,F)													//
//																	//
//---- In this array, say dependence, 								//
//---- dependence[i,j] = T,<==> file i depends on file j,			//
//---- Attention: A file is defined not to be dependent on itself	//
//																	//
//---- And in the associative array: fileNames[],					//
//---- fileNames[0] = "fileA.cpp",									//
//---- fileNames[1] = "fileB.h"; 									//
//---- fileNames[2] = "fileC.h",									//
//---- fileNames[3] = "fileD.h";									//
//////////////////////////////////////////////////////////////////////

#include "iostream.h"
#include "depMatrix.h"

// constructor
depMatrix::depMatrix()
{
	size = 0;
	dependence = NULL;
	fileNames = NULL; 
	printRequired = NULL;
}

// destructor
depMatrix::~depMatrix() 
{
	if ( dependence )
		delete dependence; 
	if ( fileNames )
		delete fileNames; 
	if ( printRequired )
		delete printRequired;
}

//return the number of the files in the matrix
int depMatrix::getSize() 
{
	return size;
}

// Return the index of a file in the dependence matrix
// If the file already exists in matrix set it returns
// index of current file, or add it as a new node in matrix
// It also return index of the file;
int depMatrix::indexOf( const char *fileName, bool isPrintFile )
{
	for ( int i=0; i<size; i++ )
	{
		if ( strcmp( fileNames[i], fileName ) == 0 ) {
			if(isPrintFile) printRequired[i] = true;
			return i;
		}
	}
	// Increment matrix
	bool *newMatrix = new bool[(size+1)*(size+1)];
	for ( i=0; i<size; i++ )
		for ( int j=0; j<size; j++ )
			newMatrix[i*(size+1)+j] = dependence[i*size+j];
		    //doesn't change the old dependence relation 

	for ( i=0; i<size+1; i++ )
	{
		//initialize the newly added colum
		newMatrix[i*(size+1)+size] = false;

		//initialize the newly added row
		newMatrix[size*(size+1)+i] = false;
	}
	delete dependence;
	dependence = newMatrix;

	// Increment file list
	char **newList = new char *[size+1];
	for ( i=0; i<size; i++ )
		newList[i] = fileNames[i];
	newList[size] = newStr(fileName );
	delete fileNames;
	fileNames = newList;

	// Increment printRequired list
	bool *newPrintRequired = new bool [size+1];
	for ( i=0; i<size; i++)
		newPrintRequired[i] = printRequired[i];
	delete printRequired;
	printRequired = newPrintRequired;

	printRequired[size] = isPrintFile;

	return size++;
}

// set one file whose index is src to be dependent on another file
// whose index is dst
void depMatrix::setDependence( int src, int dst )
{
	dependence[src*size+dst] = true;

	//if the dependence relation exist between (dst, i)
	//then the dependence relation exists between (src, i)
	for( int i=0; i<size; i++ )
		if(dependence[dst*size+i]) setDependence(src,i);

	// Use the following should be better
	//for(int i=0; i<size; i++)
	//	if(dependence[dst*size+i])
	//		for(int j=0; j<size; j++)
	//			dependence[src*size+j]
	//			= dependence[src*size+j] || dependence[i*size+j];


	//if the dependence relation exist between (i, src)
	//then the dependence relation exists between (i, dst)
	for( i=0; i<size; i++ )
		if(dependence[i*size+src]) setDependence(i,dst);

	// Use the following should be better
	// for(i=0; i<size; i++)
	//	if(dependence[i*size+src])
	//		for(int j=0; j<size; j++)
	//			dependence[i*size+dst]
	//			= dependence[i*size+dst] || dependence[i*size+j];
}

// Swap the position of the two files in dependence matrix
// and fileNames array.
void depMatrix::swap(int src, int dst)
{
	bool temp;
	char* tempName;

	// exchange the name of the two files in fileName array
	// whose index are src and dst respectively
	tempName = fileNames[src];
	fileNames[src] = fileNames[dst];
	fileNames[dst] = tempName;
	
	temp = printRequired[src];
	printRequired[src] = printRequired[dst];
	printRequired[dst] = temp;

	// exchange the include file information of the two files
	// whose index are src and dst respectively
	// namely, exchange two rows in the dependence matrix
	for( int i=0; i<size; i++) {
		temp = dependence[src*size+i];
		dependence[src*size+i] = dependence[dst*size+i];
		dependence[dst*size+i] = temp;
	}

	// exchange two columns in the dependence matrix
	for( i=0; i<size; i++) {
		temp = dependence[i*size+src];
		dependence[i*size+src] = dependence[i*size+dst];
		dependence[i*size+dst] = temp;
	}

}

// adjust the matrix so that the matrix is in decendent order
// required by the project.
// This algorithm is similar to bubble sort
// And we assume no circular dependence exist
void depMatrix::adjust()
{
	for( int i=0; i<size; i++ )
		for( int j=i+1; j<size; j++ )
			// if file[j] depends on file[i] and j > i, 
			// swap these two files
			if(dependence[j*size+i]) swap(i,j);

	//if file[i] doesn't depend on file[j]
	//sort these two files in the matrix and the array so that
	//they are placed by filename(in accending alphabet order)
	//so the header file and the implement file could be placed
	//as close as possible without violating the decendence rule
	for( i=size-1; i>=1; i--)
		for( int j=i-1; j>=0; j--)
			if( (!dependence[ j*size+i ]) && (strcmp(fileNames[j],fileNames[i]) > 0) )
				swap(i,j);
}

// Output the dependence information expressed by the
// dependence array
void depMatrix::outputDependence()
{
	// adjust the matrix before print out dependence
	adjust();

	for ( int i=0; i<size; i++ ) {		

		if(printRequired[i]) {
			cout<<"File "<<fileNames[i]<<" includes: \n";
			for ( int j=0; j<size; j++ ) {
			if ( dependence[i*size+j] )	cout << "     " << fileNames[j] <<";\n";
		}
		cout<<"\n";
		}
	}
}